-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all; 
use work.sha3_pkg.all;
use work.sha3_skein_package.all;

entity skein_control_ppl is		 
	generic (
		ppl: integer := 5;
		w: integer := 64; 
		HS : integer := HASH_SIZE_256; 
		UF : integer := 4 
	);
	port (					
		rst			: in std_logic;
		clk			: in std_logic;
		
		-- datapath signals
		-- tweak
		tw_bit_pad, tw_final, tw_first : out std_logic_vector(ppl-1 downto 0);	
		--fsm1
		ein, ec, lc			: out std_logic_vector(ppl-1 downto 0);
		zc0, zcRemNon0, zcRem1 : in std_logic_vector(ppl-1 downto 0);
		padded, final_segment : in std_logic_vector(ppl-1 downto 0);
		dth, eth, clr_rem : out std_logic_vector(ppl-1 downto 0);
		load_id, load_hash : out std_logic_vector(ppl-1 downto 0);
		
		-- pad		
		comp_rem_e0, comp_lb_e0 : in std_logic_vector(ppl-1 downto 0);
        en_lb, clr_lb : out std_logic_vector(ppl-1 downto 0);
        spos : out std_logic_vector(ppl-1 downto 0);
		
		--fsm2
		sclr_selected, shalf : out std_logic;
		er, etweak, load_tweak, snb, sf  :  out std_logic;
		slast, sfinal, lmsg, clr_msg  : out std_logic_vector(ppl-1 downto 0);
		smsg, stweak : out std_logic_vector(log2(ppl)-1 downto 0);
		-- FSM3
		shift_dout		: out std_logic_vector(ppl-1 downto 0);
		sout 		: out std_logic_vector(log2(ppl)-1 downto 0);
		sid : out std_logic;
		
		-- fifo signals
		src_ready	: in std_logic_vector(ppl-1 downto 0);
		src_read	: out std_logic_vector(ppl-1 downto 0);
		dst_ready 	: in std_logic;
		dst_write	: out std_logic
	);				 
end skein_control_ppl;

architecture struct of skein_control_ppl is				   	 
	-- fsm1
	signal block_ready_set, msg_end_set, load_next_block : std_logic_vector(ppl-1 downto 0);
	signal tw_bit_pad_en : std_logic_vector(ppl-1 downto 0);
	-- fsm2						 
		-- fsm1 communications
	signal block_ready_clr, msg_end_clr : std_logic_vector(ppl-1 downto 0); --out
	signal block_ready, msg_end : std_logic_vector(ppl-1 downto 0); --in	   
	
		-- fsm2 communications
	signal output_busy_set : std_logic_vector(ppl-1 downto 0); --out
	signal output_busy : std_logic_vector(ppl-1 downto 0); --in  
	
	-- fsm3										
	signal output_busy_clr : std_logic_vector(ppl-1 downto 0); 
	signal shift_dout_fsm3, load_hash_fsm2 : std_logic_vector(ppl-1 downto 0);
	
	-- pipelined signals 
	signal sf_temp, sclr_selected_temp : std_logic;	   
	signal etweak_sig, snb_sig : std_logic;
	signal er_sig, load_tweak_sig, shalf_sig : std_logic;
	signal lmsg_sig, clr_msg_sig : std_logic_vector(ppl-1 downto 0);
	signal dst_write_sig, sid_sig : std_logic;
	signal sfinal_sig, slast_sig, emsg_sig, sf_sig : std_logic_vector( ppl-1 downto 0);
	signal smsg_sig, stweak_sig, sout_sig : std_logic_vector(log2(ppl)-1 downto 0);
begin
	
	gfsm1: for i in 0 to ppl-1 generate
		fsm1_gen : entity work.skein_fsm1(counter) 
		port map (
			clk => clk, rst => rst, 
			zc0 => zc0(i), final_segment => final_segment(i), ein => ein(i), ec => ec(i), lc => lc(i), dth => dth(i), eth => eth(i),
			zcRemNon0 => zcRemNon0(i), zcRem1 => zcRem1(i), padded => padded(i), clr_rem => clr_rem(i), tw_bit_pad_en => tw_bit_pad_en(i), 
			load_next_block => load_next_block(i), block_ready_set => block_ready_set(i), msg_end_set => msg_end_set(i), 
			src_ready => src_ready(i), src_read => src_read(i), load_id => load_id(i),
			comp_rem_e0 => comp_rem_e0(i), comp_lb_e0 => comp_lb_e0(i), en_lb => en_lb(i), clr_lb => clr_lb(i),	spos => spos(i)
		);	 
	end generate;
	
	fsm2_gen : entity work.skein_fsm2(beh) 
	generic map ( ppl => ppl, UF => UF)	
	port map (
		clk => clk, rst => rst, 
		er => er_sig,  load_hash => load_hash_fsm2, sf => sf_sig, sfinal => sfinal_sig, snb => snb_sig, etweak => etweak_sig, slast => slast_sig,
			lmsg => lmsg_sig, load_tweak => load_tweak_sig,	stweak => stweak_sig, smsg => smsg_sig, shalf => shalf_sig, clr_msg => clr_msg_sig,
		block_ready_clr => block_ready_clr, msg_end_clr => msg_end_clr, block_ready => block_ready, msg_end => msg_end,	
		output_busy_set => output_busy_set, output_busy => output_busy
	); 
	
	fsm3_gen : entity work.sha3_fsm3_ppl(beh)
		generic map ( h => HS, w=> w, ppl => ppl )
		port map (
		clk => clk, rst => rst, 
		shift_dout => shift_dout_fsm3, sout => sout_sig,  sid => sid_sig,
		output_busy => output_busy, output_busy_clr => output_busy_clr,
		dst_ready => dst_ready, dst_write => dst_write_sig
	);	 			    				  
	
	process(clk)
	begin
		if rising_edge( clk ) then				
			sout <= sout_sig;
			sid <= sid_sig;			
			shift_dout <= shift_dout_fsm3 or load_hash_fsm2; 	
			load_hash <= load_hash_fsm2;
			dst_write <= dst_write_sig;	
			tw_final <= slast_sig or sfinal_sig;
			tw_first <= sf_sig or sfinal_sig;
			tw_bit_pad <= tw_bit_pad_en and slast_sig;			
			load_tweak <= load_tweak_sig;
			sfinal <= sfinal_sig;
			slast <= slast_sig;	
			lmsg <= lmsg_sig;	
			smsg <= smsg_sig;
			stweak <= stweak_sig;	 
			shalf <= shalf_sig;		 
			clr_msg <= clr_msg_sig;		
			sf <= sf_temp;
			sclr_selected <= sclr_selected_temp;
			snb <= snb_sig;
			etweak <= etweak_sig;
			er <= er_sig;
		end if;
	end process;
	
	-- selected	  		
	g_ppl2 : if ppl = 2 generate
		sf_temp <= sf_sig(0) when stweak_sig(0) = '0' else sf_sig(1);
		sclr_selected_temp <= clr_msg_sig(0) when smsg_sig(0) = '0' else clr_msg_sig(1);
	end generate;
	g_ppl5 : if ppl = 5 generate
		with stweak_sig(2 downto 0) select
		sf_temp <= 	sf_sig(0) when "000", sf_sig(1) when "001", sf_sig(2) when "010", 
					sf_sig(3) when "011", sf_sig(4) when others;
		with smsg_sig(2 downto 0) select
		sclr_selected_temp <= 	clr_msg_sig(0) when "000", clr_msg_sig(1) when "001", clr_msg_sig(2) when "010", 
					clr_msg_sig(3) when "011", clr_msg_sig(4) when others;
	end generate;
	g_ppl10 : if ppl = 10 generate
		with stweak_sig(3 downto 0) select
		sf_temp <= 	sf_sig( 0) when "0000", sf_sig( 1) when "0001", sf_sig( 2) when "0010", sf_sig( 3) when "0011", 
					sf_sig( 4) when "0100", sf_sig( 5) when "0101", sf_sig( 6) when "0110", sf_sig( 7) when "0111", 
					sf_sig( 8) when "1000", sf_sig( 9) when others;
		with smsg_sig(3 downto 0) select
		sclr_selected_temp <= 	clr_msg_sig( 0) when "0000", clr_msg_sig( 1) when "0001", clr_msg_sig( 2) when "0010", clr_msg_sig( 3) when "0011", 
								clr_msg_sig( 4) when "0100", clr_msg_sig( 5) when "0101", clr_msg_sig( 6) when "0110", clr_msg_sig( 7) when "0111", 
								clr_msg_sig( 8) when "1000", clr_msg_sig( 9) when others;
	end generate;
	
	comm :  for i in 0 to ppl-1 generate
		load_next_block(i) <= not block_ready(i);
		sr_blk_ready : sr_reg 
			port map ( rst => rst, clk => clk, set => block_ready_set(i), clr => block_ready_clr(i), output => block_ready(i));	
		sr_msg_end : sr_reg 
			port map ( rst => rst, clk => clk, set => msg_end_set(i), clr => msg_end_clr(i), output => msg_end(i));	
		sr_output_write : sr_reg 
			port map ( rst => rst, clk => clk, set => output_busy_set(i), clr => output_busy_clr(i), output => output_busy(i));			
	end generate; 
end struct;