-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

use work.sha3_skein_package.ALL;
use work.sha3_pkg.all;

entity skein_datapath_ppl is
	generic ( 			  
		PPL 			: integer := 10;
		VERSION 		: integer := SHA3_ROUND3;
		ADDER_TYPE 		: integer := SCCA_BASED;	  
		UF : integer 	:= 8;
		W : integer 	:= 64;
		HS : integer	:= HASH_SIZE_256 );
	port (
		-- external
		clk 			: in std_logic;
		rst 			: in std_logic;
		din  			: in std_logic_vector(ppl*w-1 downto 0);
		dout 			: out std_logic_vector(w-1 downto 0);	
		--fsm 1		
		clr_rem			: in std_logic_vector(ppl-1 downto 0);
		ein 			: in std_logic_vector(ppl-1 downto 0);	
		ec				: in std_logic_vector(ppl-1 downto 0);
		lc				: in std_logic_vector(ppl-1 downto 0);		
		dth				: in std_logic_vector(ppl-1 downto 0);		   
		eth				: in std_logic_vector(ppl-1 downto 0);		   
		zc0				: out std_logic_vector(ppl-1 downto 0);	
		zcRemNon0		: out std_logic_vector(ppl-1 downto 0);	
		zcRem1 			: out std_logic_vector(ppl-1 downto 0);	
		padded 			: out std_logic_vector(ppl-1 downto 0);	
		-- pad
		en_lb		 	: in std_logic_vector(ppl-1 downto 0); 
		clr_lb			: in std_logic_vector(ppl-1 downto 0);
        spos 			: in std_logic_vector(ppl-1 downto 0);	
		comp_rem_e0		: out std_logic_vector(ppl-1 downto 0);
		comp_lb_e0 		: out std_logic_vector(ppl-1 downto 0);
       	
		-- Tweak			  		  
		tw_bit_pad 		: in std_logic_vector(ppl-1 downto 0);	
		tw_final	 	: in std_logic_vector(ppl-1 downto 0);	
		tw_first		: in std_logic_vector(ppl-1 downto 0);	
		sfinal			: in std_logic_vector(ppl-1 downto 0); 	
		slast 			: in std_logic_vector(ppl-1 downto 0);
		
		--fsm 2	
		lmsg 			: in std_logic_vector(ppl-1 downto 0);
		clr_msg			: in std_logic_vector(ppl-1 downto 0);	
		sf				: in std_logic;	
		er				: in std_logic;	
		etweak			: in std_logic;	
		load_tweak 		: in std_logic;				
		sclr_selected 	: in std_logic;		
		snb	   			: in std_logic;	
		shalf			: in std_logic;	 
		smsg   			: in std_logic_vector(log2(ppl)-1 downto 0);
		stweak 			: in std_logic_vector(log2(ppl)-1 downto 0);
		
		-- Output				
		shift_dout		: in std_logic_vector(ppl-1 downto 0);
		load_id   		: in std_logic_vector(ppl-1 downto 0);	
		load_hash 		: in std_logic_vector(ppl-1 downto 0);
		sout 			: in std_logic_vector(log2(ppl)-1 downto 0);
		sid 			: in std_logic	
	);				  
end skein_datapath_ppl;

architecture struct of skein_datapath_ppl is 
	------ Constants	
	constant b : integer := 512; -- fixed	 
	constant nw : integer := 8; -- fixed
	constant mw			: integer := b;
	constant mwzeros	: std_logic_vector(mw-1 downto 0) := (others => '0');  
	constant log2mw : integer := log2(mw);
	constant log2mwzeros : std_logic_vector(log2mw-1 downto 0) := (others => '0');	
	constant bzeros 	: std_logic_vector(b-1 downto 0) := (others => '0');	
	constant key_size	: integer := UF/4;
	constant perm : permute_type (0 to nw-1) := get_perm( b );
	constant rot : rot_type (0 to nw/2-1,0 to 7) := get_rot( b );  
	constant iv : std_logic_vector(b-1 downto 0) := get_iv( HS, version );																			  
	constant sixteenzeros : std_logic_vector(15 downto 0) := (others => '0');	 
	constant seventyone0 : std_logic_vector(70 downto 0) := (others =>'0');
	constant remzeros : std_logic_vector(log2mw-4 downto 0) := (others => '0');
	constant remones : std_logic_vector(log2mw-4 downto 0) := (others => '1');		
	----------	
		
	---------- type declaration
	type mw_array_type is array ( 0 to ppl-1 ) of std_logic_vector(mw-1 downto 0);
	type log2mw_m4_array_type is array ( 0 to ppl-1 ) of std_logic_vector(log2mw-4 downto 0);
	type c_array_type is array ( 0 to ppl-1) of std_logic_vector(15 downto 0);
	type id_array_type is array ( 0 to ppl-1, 0 to 1) of std_logic_vector(15 downto 0);
	type tw_pos_type is array ( 0 to ppl-1 ) of std_logic_vector(21 downto 0);
	type tw_type_type is array ( 0 to ppl-1 ) of std_logic_vector(5 downto 0);
	type tweak_type is array ( 0 to ppl-1 ) of std_logic_vector(127 downto 0);
	----------	
	
	-- BYTE COUNTER 
	signal remainder, rem_data : log2mw_m4_array_type;
	signal block_count : c_array_type;
	signal cpad : std_logic_vector(ppl-1 downto 0);
	-- TWEAK
	signal tw_type : tw_type_type;
	signal tweak : tweak_type;				  
	constant tw_tree_level : std_logic_vector(6 downto 0) := (others =>'0');
	constant tw_reserved : std_logic_vector(15 downto 0) := (others =>'0');	
	constant tw_pos_zeros : std_logic_vector(73 downto 0) := (others => '0');
	signal tw_position : tw_pos_type;	   
   	
	-- Padding related signals			
	type rem_bit_type is array (0 to PPL-1) of std_logic_vector(8 downto 0);
	type ppl_8bit_type is array (0 to PPL-1) of std_logic_vector(7 downto 0);
	type ppl_wbit_type is array (0 to PPL-1) of std_logic_vector(63 downto 0);
	
	signal remaining_bits 	: rem_bit_type;
	signal din_padded 		: ppl_wbit_type;
	signal sel_din_lookup 	: ppl_8bit_type;
	signal sel_din 			: ppl_8bit_type; 
	
	-- BLOCK COUNTERS
	signal c : c_array_type;		
	
	type nw_type is array (nw-1 downto 0) of std_logic_vector(63 downto 0);		
	signal test : nw_type;
	
	type ux_nw_type is array (0 to UF) of nw_type;
	type uxm1_nw_type is array (0 to UF-1) of nw_type;
	-- ROUND SIGNALS			 
	signal min, msg : mw_array_type;
	signal min_endian1,min_endian2, msg_endian1, msg_endian2 : std_logic_vector(mw-1 downto 0);
	signal r, rmux, rp : std_logic_vector(b-1 downto 0);
	signal keyout  : round_array(key_size-1 downto 0, nw-1 downto 0);	
	signal threefish_done, cv,  keyin : std_logic_vector(b-1 downto 0);	  
	signal switch_out, switch1, switch2 : std_logic_vector(HS-1 downto 0);
	signal keyinj, keyinj_in : round_array(key_size-1 downto 0, nw-1 downto 0);	
	signal round : ux_nw_type;
	signal roundout, round_perm : uxm1_nw_type; 
	
	signal min_selected, msg_selected : std_logic_vector(mw-1 downto 0);
	signal tweak_selected : std_logic_vector(127 downto 0);
	signal shalf_reg : std_logic_vector(UF-1 downto 0);
	-- id, hash	
	constant id_size : integer := 16;
	signal id_in : std_logic_vector(ppl*id_size-1 downto 0);
begin		
	-- ///////////////////////////////////////////////////////
	-- FSM1 SECTION												
	-- block decounter	
	fsm1_section: for i in 0 to ppl-1 generate
		decounter_gen: 
			decountern 
			generic map ( N => 16, sub => 1 ) 
			port map ( clk => clk, rst => '0', load => lc(i), en => ec(i), input => din(ppl*w-(i+1)*w+log2mw+15 downto ppl*w-(i+1)*w+log2mw), output => c(i));
		
		rem1_gen : 
			decountern 
			generic map ( N => 3, sub => 1 ) 
			port map ( clk => clk, rst => clr_lb(i), load => lc(i), en => en_lb(i), 
					  input => din(8 + (PPL-i-1)*W downto 6 + (PPL-i-1)*W), 
					  output => remaining_bits(i)(8 downto 6));	
			
		rem2_reg : 
			regn generic map ( N => 6, init => "000000" ) 
			port map ( clk => clk, rst => clr_lb(i), en => lc(i), 
					   input => din(5 + (PPL-i-1)*W downto 0 + (PPL-i-1)*W), 
					   output => remaining_bits(i)(5 downto 0));						 
			
		sel_din_lookup(i) <= lookup_sel_lvl1_64bit_pad(conv_integer(remaining_bits(i)(5 downto 3)));
		sel_din(i) <= (others => '1') when spos(i) = '1' else sel_din_lookup(i);	
	
		byte_pad_gen : entity work.skein_byte_pad(struct)
			generic map ( w => 64 ) 
			port map ( din => din((PPL-i)*W-1 downto (PPL-i-1)*W), dout => din_padded(i), sel_din =>  sel_din(i));	
	
		comp_rem_e0(i) <= '1' when remaining_bits(i) = 0 else '0';
		comp_lb_e0(i) <= '1' when remaining_bits(i)(8 downto 6) = "000" else '0';
		
		remreg_gen : 
			regn 
			generic map ( N => log2mw-3, init => remzeros )
			port map ( clk => clk, rst => clr_rem(i), en => lc(i), input => din(ppl*w-(i+1)*w+log2mw-1 downto  ppl*w-(i+1)*w+3), output => remainder(i));
			
		cpad_gen: 
			process ( clk )
			begin
				if rising_edge( clk ) then
					if clr_rem(i) = '1' then
						cpad(i) <= '0';
					elsif lc(i) = '1' then
						cpad(i) <= din(2+(ppl-1-i)*w) or din(1+(ppl-1-i)*w) or din((ppl-1-i)*w);
					end if;
				end if;
			end process;

		padded(i) <= cpad(i);	
		zc0(i) <= '1' when c(i) = conv_std_logic_vector(0,16) else '0';
		zcRemNon0(i) <= '0' when remainder(i) = remzeros else '1';
		zcRem1(i) <= '1' when remainder(i) = remones else '0';	
	
		-- t gens																			  
		block_counter_gen : countern
			generic map ( N => 16 ) 
			port map ( clk => clk, rst => rst, load => dth(i), en => eth(i), input => sixteenzeros, output => block_count(i));
		
		rem_data(i) <= remainder(i) + cpad(i) when slast(i) = '1' else (others =>'0');
		tw_position(i) <= conv_std_logic_vector(8,16+log2mw-3) when sfinal(i) = '1' else block_count(i) & rem_data(i);
		
		-- tweak
		tw_type(i) <= TW_OUT_CONS when sfinal(i) = '1' else TW_MSG_CONS;
		tweak(i) <= tw_final(i) & tw_first(i) & tw_type(i) & tw_bit_pad(i) & tw_tree_level & tw_reserved & tw_pos_zeros & tw_position(i);
		
		msg_reg: 
			regn 
			generic map ( N => mw, init => mwzeros ) 
			port map ( clk => clk, rst => clr_msg(i), en => lmsg(i), input => min(i), output => msg(i) );
			
		shfin_gen : 
			sipo 
			generic map ( N => mw, M => w )
			port map (clk => clk, en => ein(i), input => din_padded(i), output => min(i) );	
	end generate;		
	-- input/output					
	
	
	min_endian1 <= switch_endian_byte(min_selected,b,64);
	min_endian2 <= switch_endian_word(min_endian1,b,64);	
	msg_endian1 <= switch_endian_byte(msg_selected,b,64);
	msg_endian2 <= switch_endian_word(msg_endian1,b,64);
		
	-----------------------
	--------- ROUND -------
	rmux <= min_endian2 when snb = '1' else rp;							
	r_reg : regn generic map ( N => b, init => bzeros ) port map ( clk => clk, rst => sclr_selected, en => er, input => rmux, output => r );
	inout_gen : for k in nw-1 downto 0 generate
		round(0)(k) <= r(iw*(k+1)-1 downto iw*k);
		rp(iw*(k+1)-1 downto iw*k) <= round(UF)(k);
	end generate;		

	-- core
	ux4_gen : if UF = 4 generate
		row_gen : for i in 0 to UF-1 generate
			key_inj : if ( i mod 4 = 0 ) generate																								
				keyinj_gen : for j in nw-1 downto 0 generate 
					add_call : adder 
						generic map ( adder_type => adder_type, n => 64 ) 
						port map ( a => keyout(i/4,j), b => round(i)(j) , s => keyinj(i/4,j));
				end generate;
				mix_gen_gen : for j in 0 to nw/2-1 generate
					mix_gen : entity work.skein_mix_4r(struct) 
						generic map ( adder_type => adder_type, rotate_0 => rot(j,i), rotate_1 => rot(j,i+4) ) 
						port map ( sel => shalf_reg(i), a => keyinj_in(i/4,2*j), b => keyinj_in(i/4,2*j+1), c => roundout(i)(2*j), d => roundout(i)(2*j+1) );
				end generate;
			end generate;
			nokey_inj : if (i mod 4 /= 0 ) generate	
				mix_gen_gen : for j in 0 to nw/2-1 generate
					mix_gen_r : entity work.skein_mix_4r(struct) 
						generic map (adder_type => adder_type,  rotate_0 => rot(j,i), rotate_1 => rot(j,i+4) ) 
						port map ( sel => shalf_reg(i), a => round(i)(2*j), b => round(i)(2*j+1), c => roundout(i)(2*j), d => roundout(i)(2*j+1) );
				end generate;
			end generate;
		end generate;	
		
		keygen_gen : entity work.skein_keygen_ppl(key_size1) 
			generic map (version => version, adder_type => adder_type, b => b, nw => nw, key_size => 1, ppl => ppl) 
			port map ( clk => clk, load => load_tweak, en => etweak, keyin => keyin, tweak => tweak_selected, keyout => keyout );
			
	end generate;

	ux8_gen : if UF = 8 generate
		row_gen : for i in 0 to UF-1 generate
			key_inj : if ( i mod 4 = 0 ) generate																								
				keyinj_gen : for j in nw-1 downto 0 generate 
					add_call : adder 
						generic map ( adder_type => adder_type, n => 64 ) 
						port map ( a => keyout(i/4,j), b => round(i)(j) , s => keyinj(i/4,j));
						--port map ( a => keyout(i/4,j), b => test(j) , s => keyinj(i/4,j));
				end generate;			
				mix_gen_gen : for j in 0 to nw/2-1 generate
					mix_gen_l : entity work.skein_mix_8r(struct) 
						generic map ( adder_type => adder_type, rotate => rot(j,i) ) 
						port map ( a => keyinj_in(i/4,2*j), b => keyinj_in(i/4,2*j+1), c => roundout(i)(2*j), d => roundout(i)(2*j+1) );	
				end generate;
			end generate;
			nokey_inj : if (i mod 4 /= 0 ) generate		
				mix_gen_gen : for j in 0 to nw/2-1 generate
					mix_gen_l : entity work.skein_mix_8r(struct) 
						generic map ( adder_type => adder_type, rotate => rot(j,i) ) 
						port map ( a => round(i)(2*j), b => round(i)(2*j+1), c => roundout(i)(2*j), d => roundout(i)(2*j+1) );	
				end generate;
			end generate;	  			
		end generate;
		
		keygen_gen : entity work.skein_keygen_ppl(key_size2_ppl10) 
			generic map (version => version, adder_type => adder_type, b => b, nw => nw, key_size => 2, ppl => ppl) 
			port map ( clk => clk, load => load_tweak, en => etweak, keyin => keyin, tweak => tweak_selected, keyout => keyout );
	end generate;
		  
		
	perm1: for i in 0 to UF-1 generate 
		perm2 : for j in 0 to nw-1 generate			
			round_perm(i)(j) <= roundout(i)(perm(j));
		end generate;
	end generate;								
	
	g_ppl5: if ppl = 5 generate
		gp: process( clk ) 
		begin
			if rising_edge( clk ) then					
				shalf_reg(0) <= shalf;
				shalf_reg(1) <= shalf_reg(0);
				shalf_reg(2) <= shalf_reg(1);			
				shalf_reg(3) <= shalf_reg(2);
				keyinj_in <= keyinj;
				round(1) <= round_perm(0);
				round(2) <= round_perm(1);
				round(3) <= round_perm(2);			
			end if;					
		end process;		
		round(4) <= round_perm(3); 
		
		with smsg(2 downto 0) select	
			min_selected <= 	min(0) when "000", min(1) when "001", min(2) when "010", min(3) when "011", min(4) when others;	  
		
		with stweak(2 downto 0) select	
			tweak_selected <= 	tweak(0) when "000", tweak(1) when "001", tweak(2) when "010", tweak(3) when "011", tweak(4) when others;
		
		with stweak(2 downto 0) select	
			msg_selected <= 	msg(0) when "000", msg(1) when "001", msg(2) when "010", msg(3) when "011", msg(4) when others;	  
	end generate;  	
	g_ppl2: if ppl = 2 generate
		gp: process( clk ) 
		begin
			if rising_edge( clk ) then									
				shalf_reg(2) <= shalf;
				round(2) <= round_perm(1);
			end if;					
		end process;										
		shalf_reg(0) <= shalf;
		shalf_reg(1) <= shalf_reg(0);
		shalf_reg(3) <= shalf_reg(2);
		keyinj_in <= keyinj;
		round(1) <= round_perm(0);
		round(3) <= round_perm(2);
		round(4) <= round_perm(3); 
		
		min_selected <= min(0) when smsg(0) = '0' else min(1);
		tweak_selected <= tweak(0) when stweak(0) = '0' else tweak(1);
		msg_selected <= msg(0) when stweak(0) = '0' else msg(1);
	end generate;				  
	g_ppl10: if ppl = 10 generate
		gp: process( clk ) 
		begin
			if rising_edge( clk ) then
				keyinj_in <= keyinj;				
				round(1) <= round_perm(0);
				round(2) <= round_perm(1);
				round(3) <= round_perm(2);
				round(4) <= round_perm(3);
				round(5) <= round_perm(4);
				round(6) <= round_perm(5);
				round(7) <= round_perm(6);
			end if;					
		end process;									
		round(8) <= round_perm(7); 
		
		with smsg(3 downto 0) select	
		min_selected <= 	min(0) when "0000", min(1) when "0001", min(2) when "0010", min(3) when "0011", 
		                    min(4) when "0100", min(5) when "0101", min(6) when "0110", min(7) when "0111", 
							min(8) when "1000", min(9) when others;	  
		
		with stweak(3 downto 0) select							 
		tweak_selected <= 	tweak(0) when "0000", tweak(1) when "0001", tweak(2) when "0010", tweak(3) when "0011", 
		                    tweak(4) when "0100", tweak(5) when "0101", tweak(6) when "0110", tweak(7) when "0111", 
							tweak(8) when "1000", tweak(9) when others;	  						
		
		with stweak(3 downto 0) select	
		msg_selected <= 	msg(0) when "0000", msg(1) when "0001", msg(2) when "0010", msg(3) when "0011", 
		                    msg(4) when "0100", msg(5) when "0101", msg(6) when "0110", msg(7) when "0111", 
							msg(8) when "1000", msg(9) when others;	  								
	end generate; 
	
	--------- ROUND -------
	-----------------------			   
	-- Key generation		
	threefish_out_gen : for i in nw-1 downto 0 generate
		threefish_done(64*(i+1)-1 downto 64*i) <= keyinj(0,i);
	end generate;
	cv <= threefish_done xor msg_endian2;
	keyin <= iv when sf = '1' else cv;						 	
	
	switch_out <= cv(HS-1 downto 0);
	switch1 <= switch_endian_word(switch_out, HS, iw); 
	switch2 <= switch_endian_byte(switch1,HS,iw);
		
	-- ///////////////////////////////////////////////////////
	-- FSM3 SECTION
	-- Round Finalization / Load new block / Output generation
	
	fsm3_output : entity work.ppl_output_unit(struct) 
		generic map ( h => HS, w => w, ppl => ppl ) 
		port map ( clk => clk, rst => rst,
					hash_result => switch2, id_in => id_in,
					sid => sid, shift_dout => shift_dout, load_hash => load_hash, 
					load_id => load_id, sout => sout,
					dout => dout );			 
					
					
	id_in_gen : for i in 0 to ppl-1 generate
		id_in((ppl-i)*id_size-1 downto (ppl-i-1)*id_size) <= din(ppl*w-(i+1)*w+15 downto ppl*w-(i+1)*w);
	end generate; 	
end struct;