-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

-- Possible generics values: 
--		HS = {HASH_SIZE_256, HASH_SIZE_512} 
--		UF = {4, 8}
--		ADDERTYPE = {SCCA_BASED, CLA_BASED, FCCA_BASED}
--		VERSION = {SHA3_ROUND2, SHA3_ROUND3}
--		PPL = {2, 5, 10}

-- Valid combinations:
--		(UF, PPL) = (4, 2), (4, 5), (8, 10)

-- Extra generic(s) :
-- 		w = {2^x} where x can be any reasonable number. By default, x is 6
-- Note : Input and output test vectors must correspond to the size of w

library ieee;
use ieee.std_logic_1164.all;
use work.sha3_skein_package.all;
use work.sha3_pkg.all;

entity skein_top_ppl is		
	generic (					 
		UF	: integer := 8;
		PPL : integer := 5;		
		ADDERTYPE : integer := SCCA_BASED; 
		W : integer := 64;		
		HS : integer := HASH_SIZE_256		
	); 
	port (		
		-- global
		rst 	: in std_logic;
		clk 	: in std_logic;	
		--fifo
		src_ready : in std_logic_vector(ppl-1 downto 0);
		src_read  : out std_logic_vector(ppl-1 downto 0);
		dst_ready : in std_logic;
		dst_write : out std_logic;		
		din		: in std_logic_vector(ppl*w-1 downto 0);
		dout	: out std_logic_vector(w-1 downto 0)
	);	   
end skein_top_ppl;


architecture structure of skein_top_ppl is   
	constant version : integer := SHA3_ROUND3;		
	-- padding unit   
	signal comp_rem_e0, comp_lb_e0, en_lb, clr_lb : std_logic_vector(ppl-1 downto 0);
	signal spos : std_logic_vector(ppl-1 downto 0);
	-- tweak
	signal tw_bit_pad, tw_final, tw_first : std_logic_vector(ppl-1 downto 0);
	-- fsm1
	signal ein, lc, ec: std_logic_vector(ppl-1 downto 0);	
	signal zc0, zcRemNon0, zcRem1, padded, clr_rem : std_logic_vector(ppl-1 downto 0);	
	signal dth, eth, final_segment : std_logic_vector(ppl-1 downto 0);
	signal sfinal, slast :  std_logic_vector(ppl-1 downto 0);
	-- fsm2
	signal shalf, er, etweak, sf : std_logic;
	signal snb, sclr_selected, load_tweak : std_logic;	
	signal smsg, stweak : std_logic_vector(log2(ppl)-1 downto 0);
	signal lmsg, clr_msg : std_logic_vector(ppl-1 downto 0);
	-- fsm3
	signal sout : std_logic_vector(log2(ppl)-1 downto 0);	
	signal sid : std_logic;
	signal shift_dout : std_logic_vector(ppl-1 downto 0);
	signal load_id, load_hash : std_logic_vector(ppl-1 downto 0);
									 	
begin		
	fseg : for i in 0 to ppl-1 generate
		final_segment(i) <= din(w*ppl-i*w-1);	-- MSB of a word
	end generate; 
	
	control_gen : entity work.skein_control_ppl(struct)  
		generic map (ppl => ppl, W=>W,  HS => HS, UF => UF )
		port map (
		rst => rst, clk => clk, 
		src_ready => src_ready, src_read => src_read, dst_ready => dst_ready, dst_write => dst_write,	  
			zc0 => zc0, zcRemNon0 => zcRemNon0, zcRem1 => zcRem1, padded => padded, clr_rem => clr_rem, 
			tw_bit_pad => tw_bit_pad, tw_final => tw_final, tw_first => tw_first,
		final_segment => final_segment, ein => ein, ec => ec, lc => lc, dth => dth, eth => eth, 
		-- fsm 2
		er => er, sf => sf, sfinal => sfinal, sclr_selected => sclr_selected, clr_msg => clr_msg, 
			snb => snb,  shalf => shalf, slast => slast, etweak => etweak,
			lmsg => lmsg, load_tweak => load_tweak, smsg => smsg, stweak => stweak, 
		-- fsm 3
		shift_dout => shift_dout, load_id => load_id, load_hash => load_hash,  sout => sout,  sid => sid,
		 -- pad
        comp_rem_e0 => comp_rem_e0, comp_lb_e0 => comp_lb_e0, en_lb => en_lb, clr_lb => clr_lb,	spos => spos
	);			
	
	datapath_gen : entity work.skein_datapath_ppl(struct) 
		generic map (ppl => ppl, version => version, adder_type => ADDERTYPE, w => w, HS => HS, UF => UF  )
		port map (
		rst => rst, clk => clk, din => din, dout => dout,
		ein => ein, ec => ec, lc => lc, dth => dth, eth => eth,
			zc0 => zc0, zcRemNon0 => zcRemNon0, zcRem1 => zcRem1, padded => padded, clr_rem => clr_rem, 
			tw_bit_pad => tw_bit_pad, tw_final => tw_final, tw_first => tw_first,
		-- fsm 2
		er => er, sf => sf, sfinal => sfinal, sclr_selected => sclr_selected, clr_msg => clr_msg,
		    snb => snb,  shalf => shalf, slast => slast, etweak => etweak,
			lmsg => lmsg, load_tweak => load_tweak, smsg => smsg, stweak => stweak, 
		-- fsm 3
		shift_dout => shift_dout, load_id => load_id, load_hash => load_hash,  sout => sout,  sid => sid,
		 -- pad
        comp_rem_e0 => comp_rem_e0, comp_lb_e0 => comp_lb_e0, en_lb => en_lb, clr_lb => clr_lb,	spos => spos
	);

end structure;
	
	