-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all; 
use ieee.std_logic_arith.all;
use work.sha3_pkg.all;
use work.sha3_skein_package.all;

entity skein_keygen_ppl is	
	generic ( 	ppl : integer := 5;
				version : integer := SHA3_ROUND3;
				adder_type : integer := SCCA_BASED;
				b : integer := 512; 
				key_size : integer := 1;
				nw : integer := 8
				);
	port (					
		clk		: in std_logic;
		load	: in std_logic;
		en		: in std_logic;
		keyin	: in std_logic_vector(b-1 downto 0);
		tweak 	: in std_logic_vector(127 downto 0);
		keyout	: out round_array(key_size-1 downto 0, nw-1 downto 0)
	);
end skein_keygen_ppl;
		
architecture key_size1 of skein_keygen_ppl is
	constant key_const : std_logic_vector(63 downto 0) := get_key_const( version );
	-- tweak											
	type tweak_type is array (2 downto 0) of std_logic_vector(63 downto 0);	
	signal tw_iv, tw_reg, tw_sel, tw_ppl : tweak_type;
	
	-- key
	type nw_type is array (nw-1 downto 0) of std_logic_vector(63 downto 0);
	
	signal parkey   : nw_type;
	signal parkey_in : std_logic_vector(63 downto 0);
	
	-- output of mux
	signal key_sel : nw_type; 
	signal key_ppl : nw_type;											  
	
	type key_reg_type is array ( nw-1 downto nw-4) of std_logic_vector(63 downto 0);
	signal key_reg : key_reg_type; -- only using the top 3 (nw-1 downto nw-3) 
	
	-- computed values
	signal key_computed : nw_type;
	-- keyregout
	signal key_out_reg : nw_type;
	
	-- sub key counter		
	signal s, s_add, s_reg, s_ppl : std_logic_vector(6 downto 0);
	signal s_reg_extended : std_logic_vector(63 downto 0);		
begin					  
	----------------------
	-- tweak
		--tweak init	   
	tw_iv(2) <= tweak(127 downto 64) xor tweak(63 downto 0);
	tw_iv(1) <= tweak(127 downto 64);
	tw_iv(0) <= tweak(63 downto 0);
	
		--												   
	g_tw_reg : for i in 0 to 2 generate
		tw_sel(i) <= tw_iv(i) when load = '1' else tw_reg(i);	  
		
		g_shfreg : shfreg 
			generic map ( width => iw, depth => ppl-1 ) 
			port map ( clk => clk, en => '1', i => tw_sel(i), o => tw_ppl(i) );		
			
		g_reg : regn 
			generic map ( N => iw, init => iwzeros ) 
			port map ( clk => clk, rst => '0', en => en, input => tw_ppl((i+1) mod 3), output => tw_reg(i) );
	end generate;
							
	
	----------------------
	-- Subkey counter	
	s <= (others => '0') when load = '1' else s_reg;
	
	s_add <= s_ppl + 1;		
	sreg_gen : regn 
			generic map ( N => 7, init => "0000000" ) 
			port map ( clk => clk, rst => '0', en => en, input => s_add, output => s_reg );
			
	g_s_ppl_reg : shfreg 
			generic map ( width => 7, depth => ppl-1 ) 
			port map ( clk => clk, en => '1', i => s, o => s_ppl );		
	----------------------
	-- key
	g_key_top : for i in nw-1 downto nw-4 generate
		key_sel(i) <= keyin(iw*i+iw-1 downto i*iw) when load = '1' else key_reg(i);		
		g_key_top_shf : shfreg 
			generic map ( width => iw, depth => ppl-1 ) 
			port map ( clk => clk, en => '1', i => key_sel(i), o => key_ppl(i) );	
	end generate;								  									
	
	key_reg0 : regn 
			generic map ( N => iw, init => iwzeros ) 
			port map ( clk => clk, rst => '0', en => en, input => parkey_in, output => key_reg(nw-1));
	key_reg1 : regn 
			generic map ( N => iw, init => iwzeros ) 
			port map ( clk => clk, rst => '0', en => en, input => key_ppl(nw-1), output => key_reg(nw-2));
	key_reg2 : regn 
			generic map ( N => iw, init => iwzeros ) 
			port map ( clk => clk, rst => '0', en => en, input => key_ppl(nw-2), output => key_reg(nw-3));	
	key_reg3 : regn 
			generic map ( N => iw, init => iwzeros ) 
			port map ( clk => clk, rst => '0', en => en, input => key_ppl(nw-3), output => key_reg(nw-4));			
	
	g_key_bot : for i in nw-5 downto 0 generate
		key_sel(i) <= keyin(iw*i+iw-1 downto i*iw) when load = '1' else key_out_reg(i+1);		
		g_key_top_shf : shfreg 
			generic map ( width => iw, depth => ppl-1 ) 
			port map ( clk => clk, en => '1', i => key_sel(i), o => key_ppl(i) );	
	end generate;
	
	-- parity key 
	parkey(0) <= key_ppl(0);
	parkey_gen : for j in 1 to nw-1 generate
		parkey(j) <= key_ppl(j) xor parkey(j-1);
	end generate;
	parkey_in <= parkey(nw-1) xor key_const;
	
	s_reg_extended <= conv_std_logic_vector(0,57) & s_ppl;
	add_call1 : adder generic map ( adder_type => adder_type, n => 64 ) port map ( a => key_ppl(nw-1), b => s_reg_extended, s => key_computed(nw-1));
	add_call2 : adder generic map ( adder_type => adder_type, n => 64 ) port map ( a => key_ppl(nw-2), b => tw_ppl(1), s => key_computed(nw-2));
	add_call3 : adder generic map ( adder_type => adder_type, n => 64 ) port map ( a => key_ppl(nw-3), b => tw_ppl(0), s => key_computed(nw-3));
	gkey: for i in nw-4 downto 0 generate
		key_computed(i) <= key_ppl(i);
	end generate;
	
	output_proc : process( clk ) 
	begin
		if rising_edge(clk) then
			key_out_reg <= key_computed;
		end if;
	end process;	
	
	gout: for i in nw-1 downto 0 generate
		keyout(0,i) <= key_out_reg(i);
	end generate;
end key_size1;

architecture key_size2_ppl10 of skein_keygen_ppl is
	constant key_const : std_logic_vector(63 downto 0) := get_key_const( version );
	-- tweak											
	type tweak_type is array (2 downto 0) of std_logic_vector(63 downto 0);	
	signal tw_iv, tw_reg, tw_sel, tw_ppl, tw_ppl2 : tweak_type;
	
	-- key
	type nw_type is array (nw-1 downto 0) of std_logic_vector(63 downto 0);
	
	signal parkey, parkey2   : nw_type;
	signal parkey_in, parkey_in2 : std_logic_vector(63 downto 0);
	
	-- output of mux
	signal key_sel : nw_type; 
	signal key_ppl, key_ppl3 : nw_type;											  
	
	type key_reg_type is array ( nw-1 downto nw-4) of std_logic_vector(63 downto 0);
	signal key_reg : key_reg_type; -- only using the top 3 (nw-1 downto nw-3) 
	
	-- computed values	
	type nw_type2 is array (nw-1 downto nw-3) of std_logic_vector(63 downto 0);
	signal key_computed, key_computed2 : nw_type2;
	-- keyregout
	signal key_out_reg, key_out_reg2 : nw_type;	 
	
	
	signal d1_1, d1_2, d1_3, d1_4, d1_5 : nw_type;
	signal d2_1, d2_2, d2_3, d2_4, d2_5 : nw_type;
	
	-- sub key counter		
	signal s, s_add, s_add2, s_reg, s_ppl, s_ppl2 : std_logic_vector(6 downto 0);
	signal s_reg_extended,s_reg_extended2 : std_logic_vector(63 downto 0);		
begin					  
	----------------------
	-- tweak
		--tweak init	   
	tw_iv(2) <= tweak(127 downto 64) xor tweak(63 downto 0);
	tw_iv(1) <= tweak(127 downto 64);
	tw_iv(0) <= tweak(63 downto 0);
	
		--												   
	g_tw_reg : for i in 0 to 2 generate
		tw_sel(i) <= tw_iv(i) when load = '1' else tw_reg(i);	  
		
		g_shfreg : shfreg 
			generic map ( width => iw, depth => 4 ) 
			port map ( clk => clk, en => '1', i => tw_sel(i), o => tw_ppl(i) );	
		g_shfreg2 : shfreg 
			generic map ( width => iw, depth => 5 ) 
			port map ( clk => clk, en => '1', i => tw_ppl((i+1) mod 3), o => tw_ppl2(i) );	
			
		g_reg : regn 
			generic map ( N => iw, init => iwzeros ) 
			port map ( clk => clk, rst => '0', en => en, input => tw_ppl2((i+1) mod 3), output => tw_reg(i) );
	end generate;
							
	
	----------------------
	-- Subkey counter	
	s <= (others => '0') when load = '1' else s_reg;
	
	s_add2 <= s_ppl2 + 1;
	s_add <= s_ppl + 1;		
	sreg_gen : regn 
			generic map ( N => 7, init => "0000000" ) 
			port map ( clk => clk, rst => '0', en => en, input => s_add2, output => s_reg );
			
	g_s_ppl_reg : shfreg 
			generic map ( width => 7, depth => 4 ) 
			port map ( clk => clk, en => '1', i => s, o => s_ppl );			   
	g_s_ppl_reg2 : shfreg 
			generic map ( width => 7, depth => 5) 
			port map ( clk => clk, en => '1', i => s_add, o => s_ppl2 );
	----------------------
	-- key
	g_key_top : for i in nw-1 downto nw-4 generate
		key_sel(i) <= keyin(iw*i+iw-1 downto i*iw) when load = '1' else key_reg(i);		
		g_key_top_shf : shfreg 
			generic map ( width => iw, depth => 4 ) 
			port map ( clk => clk, en => '1', i => key_sel(i), o => key_ppl(i) );		
			
		nw1: if i = nw-1 generate
			g_key_top_shf2 : shfreg 
				generic map ( width => iw, depth => 5 ) 
				port map ( clk => clk, en => '1', i => parkey_in, o => key_ppl3(i) );
		end generate;
		nwx: if i /= nw-1 generate
			g_key_top_shf2 : shfreg 
				generic map ( width => iw, depth => 5 ) 
				port map ( clk => clk, en => '1', i => key_ppl(i+1), o => key_ppl3(i) );
		end generate;
	end generate;								  									
	
	key_reg0 : regn 
			generic map ( N => iw, init => iwzeros ) 
			port map ( clk => clk, rst => '0', en => en, input => parkey_in2, output => key_reg(nw-1));
	key_reg1 : regn 
			generic map ( N => iw, init => iwzeros ) 
			port map ( clk => clk, rst => '0', en => en, input => key_ppl3(nw-1), output => key_reg(nw-2));
	key_reg2 : regn 
			generic map ( N => iw, init => iwzeros ) 
			port map ( clk => clk, rst => '0', en => en, input => key_ppl3(nw-2), output => key_reg(nw-3));	
	key_reg3 : regn 
			generic map ( N => iw, init => iwzeros ) 
			port map ( clk => clk, rst => '0', en => en, input => key_ppl3(nw-3), output => key_reg(nw-4));			
	
	g_key_bot : for i in nw-5 downto 0 generate
		key_sel(i) <= keyin(iw*i+iw-1 downto i*iw) when load = '1' else key_out_reg2(i+1);		
		g_key_top_shf : shfreg 
			generic map ( width => iw, depth => 4 ) 
			port map ( clk => clk, en => '1', i => key_sel(i), o => key_ppl(i) );	
		g_key_top_shf3 : shfreg 
			generic map ( width => iw, depth => 5 ) 
			port map ( clk => clk, en => '1', i => key_ppl(i+1), o => key_ppl3(i) );	
	end generate;
	
	-- parity key 
	parkey(0) <= key_ppl(0);   
	parkey2(0) <= key_ppl3(0);
	parkey_gen : for j in 1 to nw-1 generate
		parkey(j) <= key_ppl(j) xor parkey(j-1);
		parkey2(j) <= key_ppl3(j) xor parkey2(j-1);
	end generate;
	parkey_in <= parkey(nw-1) xor key_const;
	parkey_in2 <= parkey2(nw-1) xor key_const;
	
	s_reg_extended <= conv_std_logic_vector(0,57) & s_ppl;	
	s_reg_extended2 <= conv_std_logic_vector(0,57) & s_ppl2;
	add_call1 : adder generic map ( adder_type => adder_type, n => 64 ) port map ( a => key_ppl(nw-1), b => s_reg_extended, s => key_computed(nw-1));
	add_call2 : adder generic map ( adder_type => adder_type, n => 64 ) port map ( a => key_ppl(nw-2), b => tw_ppl(1), s => key_computed(nw-2));
	add_call3 : adder generic map ( adder_type => adder_type, n => 64 ) port map ( a => key_ppl(nw-3), b => tw_ppl(0), s => key_computed(nw-3));
	add_call4 : adder generic map ( adder_type => adder_type, n => 64 ) port map ( a => key_ppl3(nw-1), b => s_reg_extended2, s => key_computed2(nw-1));
	add_call5 : adder generic map ( adder_type => adder_type, n => 64 ) port map ( a => key_ppl3(nw-2), b => tw_ppl2(1), s => key_computed2(nw-2));
	add_call6 : adder generic map ( adder_type => adder_type, n => 64 ) port map ( a => key_ppl3(nw-3), b => tw_ppl2(0), s => key_computed2(nw-3));
	
	output_proc : process( clk ) 
	begin
		if rising_edge(clk) then 
			for i in nw-1 downto nw-3 loop
				key_out_reg(i) <= key_computed(i);
				key_out_reg2(i) <= key_computed2(i);
				
			end loop;
			for i in nw-4 downto 0 loop			
				key_out_reg(i) <= key_ppl(i);
				key_out_reg2(i) <= key_ppl3(i);
			end loop;
		end if;
	end process;	  	
		
	-- outgen : for i in nw-4 downto 1 generate
		-- key_out_reg(i) <= key_ppl2(i-1);
	-- end generate;				
	
	dd: process( clk )
	begin
		if rising_edge(clk) then
			d1_1 <= key_out_reg;
			d1_2 <= d1_1;
			d1_3 <= d1_2;
			d1_4 <= d1_3;
			d1_5 <= d1_4;
			
			d2_1 <= key_out_reg2;
			d2_2 <= d2_1;
			d2_3 <= d2_2;
			d2_4 <= d2_3;
			d2_5 <= d2_4;  					
		end if;
	end process;
	
	
	gout: for i in nw-1 downto 0 generate
		keyout(0,i) <= d1_5(i);
		keyout(1,i) <= d2_5(i);
	end generate;
end key_size2_ppl10;